/**
 * <h1>AssignmentGenerator</h1>
 *
 * <p>Generate code for an assignment statement.</p>
 *
 * <p>Copyright (c) 2017 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#include <string>
#include <vector>
#include "AssignmentGenerator.h"
#include "StatementGenerator.h"
#include "ExpressionGenerator.h"
#include "../Instruction.h"
#include "../../../Object.h"
#include "../../../frontend/Token.h"
#include "../../../frontend/pascal/PascalToken.h"
#include "../../../intermediate/SymTabEntry.h"
#include "../../../intermediate/symtabimpl/SymTabEntryImpl.h"
#include "../../../intermediate/symtabimpl/Predefined.h"
#include "../../../intermediate/ICodeNode.h"
#include "../../../intermediate/icodeimpl/ICodeNodeImpl.h"
#include "../../../intermediate/TypeSpec.h"
#include "../../../intermediate/typeimpl/TypeSpecImpl.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::intermediate;
using namespace wci::intermediate::symtabimpl;
using namespace wci::intermediate::icodeimpl;
using namespace wci::intermediate::typeimpl;
using namespace wci::backend::compiler;

AssignmentGenerator::AssignmentGenerator(CodeGenerator *parent)
    : StatementGenerator(parent)
{
}

void AssignmentGenerator::generate(ICodeNode *node) throw (string)
{
    TypeSpec *assign_typespec = node->get_typespec();

    // The ASSIGN node's children are the target variable
    // and the expression.
    vector<ICodeNode *> *assign_children = node->get_children();
    ICodeNode *target_node = (*assign_children)[0];
    ICodeNode *expr_node = (*assign_children)[1];

    SymTabEntry *target_id = target_node->get_id();
    TypeSpec *target_typespec = target_node->get_typespec();
    TypeSpec *expr_typespec = expr_node->get_typespec();
    ExpressionGenerator expr_generator(this);

    int slot;           // local variables array slot number of the target
    int nesting_level;  // nesting level of the target
    SymTab *symtab;     // symbol table that contains the target id

    // Assign a function value. Use the slot number of the function value.
    if (target_id->get_definition() == (Definition) DF_FUNCTION)
    {
        slot = target_id->get_slot_number();
        nesting_level = 2;
    }

    // Standard assignment.
    else
    {
        symtab = target_id->get_symtab();
        slot = target_id->get_slot_number();
        nesting_level = symtab->get_nesting_level();
    }

    // Assign to a VAR parameter: Load the address of the wrapper.
    if (is_wrapped(target_id)) {
        emit_load_local(nullptr, slot);
        local_stack->increase(1);
    }

    vector<ICodeNode *> *target_children = target_node->get_children();
    int children_count = target_children->size();

    // The last modifier, if any, is the variable's last subscript or field.
    ICodeNode *last_modifier = nullptr;

    // The target variable has subscripts and/or fields.
    if (children_count > 0)
    {
        ExpressionGenerator expr_generator(this);
        last_modifier = (*target_children)[children_count-1];

        if (assign_typespec->is_pascal_string())
        {
            expr_generator.generate_load_value(target_node);
        }
        else
        {
            assign_typespec =
                    expr_generator.generate_load_variable(target_node);
        }
    }

    // Assign to a Pascal string.
    else if (assign_typespec->is_pascal_string())
    {
        emit_load_variable(target_id);
        local_stack->increase(1);
    }

    // Generate code to do the assignment.
    if (target_typespec->is_pascal_string())
    {
        generate_string_assignment(assign_typespec, expr_node,
                                   expr_typespec, expr_generator);
    }
    else
    {
        generate_scalar_assignment(target_typespec, target_id,
                                   last_modifier, slot, nesting_level,
                                   expr_node, expr_typespec, expr_generator);
    }
}

void AssignmentGenerator::generate_scalar_assignment(
                                    TypeSpec *target_typespec,
                                    SymTabEntry *target_id,
                                    ICodeNode *last_modifier,
                                    int index,
                                    int nesting_level,
                                    ICodeNode *expr_node,
                                    TypeSpec *expr_typespec,
                                    ExpressionGenerator &expr_generator)
{
    // Generate code to evaluate the expression.
    // Special cases: float variable := integer constant
    //                float variable := integer expression
    //                char variable  := single-character string constant
    if (target_typespec == Predefined::real_type)
    {
        if (expr_node->get_type() == (ICodeNodeType) NT_INTEGER_CONSTANT)
        {
            int value = cast((*(expr_node->get_value())), int);
            emit_load_constant((float) value);
            local_stack->increase(1);
        }
        else
        {
            expr_generator.generate(expr_node);
            if (expr_typespec->base_type() == Predefined::integer_type)
            {
                emit(I2F);
            }
        }
    }
    else if (   (target_typespec == Predefined::char_type)
             && (expr_node->get_type()
                                 == (ICodeNodeType) NT_STRING_CONSTANT))
    {
        int value = cast((*(expr_node->get_value())), string)[0];
        emit_load_constant(value);
        local_stack->increase(1);
    }
    else
    {
        expr_generator.generate(expr_node);
    }

    // Generate code to store the expression value into the target variable.
    // The target variable has no subscripts or fields.
    if (last_modifier == nullptr)
    {
        emit_store_variable(target_id, nesting_level, index);
        local_stack->decrease(is_wrapped(target_id) ? 2 : 1);
    }

    // The target variable is a field.
    else if (last_modifier->get_type() == (ICodeNodeType) NT_FIELD)
    {
        TypeSpec *data_typespec = last_modifier->get_typespec()->base_type();
        TypeForm type_form = data_typespec->get_form();

        if (   (type_form == (TypeForm) TF_SCALAR)
            || (type_form == (TypeForm) TF_ENUMERATION))
        {
            emit(INVOKESTATIC, value_of_signature(data_typespec));
        }

        emit(INVOKEVIRTUAL,
             string("java/util/HashMap.put(Ljava/lang/Object;") +
             string("Ljava/lang/Object;)Ljava/lang/Object;"));
        emit(POP);
        local_stack->decrease(3);
    }

    // The target variable is an array element.
    else {
        emit_store_array_element(target_typespec);
        local_stack->decrease(3);
    }
}

const string AssignmentGenerator::SETLENGTH =
        "java/lang/StringBuilder.setLength(I)V";
const string AssignmentGenerator::PAD_BLANKS =
        "PaddedString.blanks(II)Ljava/lang/StringBuilder;";

const string AssignmentGenerator::APPEND_STRING =
        string("java/lang/StringBuilder.append(Ljava/lang/String;)") +
        string("Ljava/lang/StringBuilder;");
const string AssignmentGenerator::APPEND_CHARSEQUENCE =
        string("java/lang/StringBuilder.append(Ljava/lang/CharSequence;)") +
        string("Ljava/lang/StringBuilder;");

const string AssignmentGenerator::STRINGBUILDER_SUBSTRING =
        "java/lang/StringBuilder.substring(II)Ljava/lang/String;";
const string AssignmentGenerator::STRING_SUBSTRING =
        "java/lang/String.substring(II)Ljava/lang/String;";

void AssignmentGenerator::generate_string_assignment(
                                    TypeSpec *target_typespec,
                                    ICodeNode *expr_node,
                                    TypeSpec *expr_typespec,
                                    ExpressionGenerator& expr_generator)
{
    int target_length = target_typespec->get_array_elmt_count();
    int source_length;
    string appender;

    emit(DUP);
    emit_load_constant(0);
    emit(INVOKEVIRTUAL, SETLENGTH);

    local_stack->use(2, 2);

    // Generate code to load the source string.
    if (expr_node->get_type() == (ICodeNodeType) NT_STRING_CONSTANT)
    {
        string value = cast((*(expr_node->get_value())), string);
        source_length = value.length();
        appender = APPEND_STRING;
        emit_load_constant(value);

        local_stack->increase(1);
    }
    else
    {
        source_length = expr_typespec->get_array_elmt_count();
        appender = APPEND_CHARSEQUENCE;
        expr_generator.generate(expr_node);
    }

    // Same lengths.
    if (target_length == source_length) {
        emit(INVOKEVIRTUAL, appender);
        local_stack->decrease(1);
    }

    // Truncate if necessary.
    else if (target_length < source_length) {
        emit_load_constant(0);
        emit_load_constant(target_length);

        string substring_routine =
            expr_node->get_type() == (ICodeNodeType) NT_STRING_CONSTANT
                                          ? STRING_SUBSTRING
                                          : STRINGBUILDER_SUBSTRING;

        emit(INVOKEVIRTUAL, substring_routine);
        emit(INVOKEVIRTUAL, appender);

        local_stack->use(2, 3);
    }

    // Blank-pad if necessary.
    else {
        emit(INVOKEVIRTUAL, appender);
        emit_load_constant(target_length);
        emit_load_constant(source_length);
        emit(INVOKESTATIC,PAD_BLANKS);
        emit(INVOKEVIRTUAL, APPEND_CHARSEQUENCE);

        local_stack->use(2, 3);
    }

    emit(POP);
    local_stack->decrease(1);
}

}}}}  // namespace wci::backend::compiler::generators
