/**
 * <h1>CallDeclaredGenerator</h1>
 *
 * <p>Generate code for a call to a declared routine.</p>
 *
 * <p>Copyright (c) 2018 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#include <string>
#include <vector>
#include "CallDeclaredGenerator.h"
#include "CallGenerator.h"
#include "ExpressionGenerator.h"
#include "../CodeGenerator.h"
#include "../../../intermediate/SymTabEntry.h"
#include "../../../intermediate/symtabimpl/SymTabEntryImpl.h"
#include "../../../intermediate/symtabimpl/Predefined.h"
#include "../../../intermediate/ICodeNode.h"
#include "../../../intermediate/icodeimpl/ICodeNodeImpl.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::intermediate;
using namespace wci::intermediate::symtabimpl;
using namespace wci::intermediate::icodeimpl;
using namespace wci::backend::compiler;

CallDeclaredGenerator::CallDeclaredGenerator(CodeGenerator *parent)
    : CallGenerator(parent)
{
}

void CallDeclaredGenerator::generate(ICodeNode *node)
{
    // Generate code for any actual parameters.
    if (node->get_children()->size() > 0)
    {
        generate_actual_parms(node);
    }

    // Generate code to make the call.
    generate_call(node);

    // Generate code for the epilogue.
    if (node->get_children()->size() > 0)
    {
        generate_call_epilogue(node);
    }

    // A function call leaves a value on the operand stack.
    SymTabEntry *routine_id = node->get_id();
    if (routine_id->get_definition() == (Definition) DF_FUNCTION)
    {
        local_stack->increase(1);
    }
}

void CallDeclaredGenerator::generate_actual_parms(ICodeNode *call_node)
{
    SymTabEntry *routine_id = call_node->get_id();
    ICodeNode *parms_node = (*call_node->get_children())[0];
    vector<ICodeNode *> *actual_nodes = parms_node->get_children();
    vector<SymTabEntry *> *formal_ids = routine_id->get_routine_parms();

    // Iterate over the formal parameters.
    for (int i = 0; i < formal_ids->size(); ++i)
    {
        SymTabEntry *formal_id = (*formal_ids)[i];
        ICodeNode *actual_node = (*actual_nodes)[i];
        TypeSpec *formal_typespec = formal_id->get_typespec();
        TypeSpec *actual_typespec = actual_node->get_typespec();

        // VAR parameter: An actual parameter that is not structured
        //                needs to be wrapped.
        if (is_wrapped(formal_id))
        {
            SymTabEntry *actual_id = actual_node->get_id();
            int wrap_slot = actual_id->get_wrap_slot_number();

            // Already wrapped: Load the wrapper.
            if (wrap_slot != -1)
            {
                emit_load_local(nullptr, wrap_slot);
                local_stack->increase(1);
            }

            // Actual parameter is itself a VAR parameter: No further
            //                                             wrapping.
            else if (actual_id->get_definition()
                                            == (Definition) DF_VAR_PARM)
            {
                int actual_slot = actual_id->get_slot_number();
                emit_load_local(nullptr, actual_slot);
                local_stack->increase(1);
            }

            // Need to wrap: Reserve a temporary variable to hold the
            //               wrapper's address.
            else
            {
                ExpressionGenerator expr_generator(this);
                int wrap_slot = local_variables->reserve();
                actual_id->set_wrap_slot_number(wrap_slot);
                generate_wrap(actual_node, formal_typespec, wrap_slot,
                              expr_generator);
            }
        }

        // Value parameter: Actual parameter is a constant string.
        else if (   (formal_typespec == Predefined::char_type)
                 && (actual_node->get_type()
                                == (ICodeNodeType) NT_STRING_CONSTANT))
        {
            string str = cast((*(actual_node->get_value())), string);

            // Constant character.
            if (formal_typespec == Predefined::char_type)
            {
                int value = str[0];
                emit_load_constant(value);
            }

            // Constant string.
            else
            {
                emit_load_constant((int) str.length());
                emit(INVOKESTATIC,
                     "PaddedString.create(I)Ljava/lang/StringBuilder;");
                emit_load_constant(str);
                emit(INVOKEVIRTUAL,
                     string("java/lang/StringBuilder.append(Ljava/lang/String;)") +
                     string("Ljava/lang/StringBuilder;"));
            }

            local_stack->increase(1);
        }

        // Value parameter: All other types.
        else
        {
            ExpressionGenerator expr_generator(this);
            expr_generator.generate(actual_node);
            emit_range_check(formal_typespec);

            // real formal := integer actual
            if (   (formal_typespec == Predefined::real_type)
                && (actual_typespec->base_type()
                                            == Predefined::integer_type))
            {
                emit(I2F);
            }

            // Structured data needs to be cloned.
            else if (needs_cloning(formal_id))
            {
                clone_actual_parameter(formal_typespec);
            }
        }
    }
}

void CallDeclaredGenerator::generate_wrap(
                    ICodeNode *actual_node, TypeSpec *formal_typespec,
                    int wrap_slot, ExpressionGenerator& expr_generator)
{
    // Wrap the value of an actual parameter.
    string wrapper = var_parm_wrapper(formal_typespec);  // selected wrapper

    // Create the wrapper.
    emit(NEW, wrapper);
    emit(DUP);
    local_stack->increase(2);

    // Generate code to evaluate the actual parameter value.
    expr_generator.generate(actual_node);

    // Invoke the wrapper's constructor to wrap the parameter value.
    string init = wrapper + "/<init>(" + type_descriptor(formal_typespec)
                          + ")V";
    emit(INVOKENONVIRTUAL, init);
    local_stack->decrease(1);

    // Store wrapper's address into a temporary variable and
    // leave a copy on the operand stack.
    emit(DUP);
    emit_store_local(nullptr, wrap_slot);
    local_stack->use(1);
}

void CallDeclaredGenerator::clone_actual_parameter(
                                        TypeSpec *formal_typespec)
{
    emit(INVOKESTATIC, string("Cloner.deepClone(Ljava/lang/Object;)") +
                       string("Ljava/lang/Object;"));
    emit_check_cast(formal_typespec);
}

void CallDeclaredGenerator::generate_call(ICodeNode *call_node)
{
    SymTabEntry *routine_id = call_node->get_id();
    string *routine_name = routine_id->get_name();
    string buffer = "";

    // Procedure or function name.
    buffer += program_name;
    buffer += "/";
    buffer += *routine_name;
    buffer += "(";

    // Parameter and return type descriptors.
    vector<SymTabEntry *> *parm_ids = routine_id->get_routine_parms();
    if (parm_ids != nullptr)
    {
        for (SymTabEntry *parm_id : *parm_ids)
        {
            buffer += type_descriptor(parm_id);
        }
    }
    buffer += ")";
    buffer += type_descriptor(routine_id);

    // Generate a call to the routine.
    emit(INVOKESTATIC, buffer);

    if (parm_ids != nullptr)
    {
        local_stack->decrease(parm_ids->size());
    }
}

void CallDeclaredGenerator::generate_call_epilogue(ICodeNode *call_node)
{
    SymTabEntry *routine_id = call_node->get_id();
    ICodeNode *parms_node = (*call_node->get_children())[0];
    vector<ICodeNode *> *actual_nodes = parms_node->get_children();
    vector<SymTabEntry *> *formal_ids = routine_id->get_routine_parms();

    // Iterate over the formal parameters.
    for (int i = 0; i < formal_ids->size(); ++i)
    {
        SymTabEntry *formal_id = (*formal_ids)[i];
        ICodeNode *actual_node = (*actual_nodes)[i];
        TypeSpec *formal_typespec = formal_id->get_typespec();

        // Wrapped parameters only.
        if (is_wrapped(formal_id))
        {
            SymTabEntry *actual_id = actual_node->get_id();

            // If the actual parameter is itself a VAR parameter,
            // keep it wrapped. Otherwise unwrap its value.
            if (actual_id->get_definition() != (Definition) DF_VAR_PARM)
            {
                generate_unwrap(actual_id, formal_typespec, program_name);
            }
        }
    }
}

void CallDeclaredGenerator::generate_unwrap(SymTabEntry *actual_id, 
                                            TypeSpec *formal_typespec, 
                                            string program_name)
{
    SymTab *symtab = actual_id->get_symtab();
    int actual_slot = actual_id->get_slot_number();
    int wrap_slot = actual_id->get_wrap_slot_number();
    string type_desc = type_descriptor(formal_typespec);
    int nesting_level = symtab->get_nesting_level();
    string wrapper = var_parm_wrapper(formal_typespec);  // selected wrapper

    // Load the wrapper and get its value.
    emit_load_local(nullptr, wrap_slot);
    emit(GETFIELD, wrapper + "/value", type_desc);

    // Store the value back into the original variable.
    if (nesting_level == 1)
    {
        string actual_name = program_name + "/" + *(actual_id->get_name());
        emit(PUTSTATIC, actual_name, type_desc);
    }
    else
    {
        emit_store_local(formal_typespec, actual_slot);
    }

    local_stack->use(1, 2);
}

}}}}  // namespace wci::backend::compiler::generators
