/**
 * <h1>CallStandardGenerator</h1>
 *
 * <p>Generate code for a call to a standard routine.</p>
 *
 * <p>Copyright (c) 2018 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */

#include <string>
#include <vector>
#include "CallStandardGenerator.h"
#include "CallGenerator.h"
#include "ExpressionGenerator.h"
#include "../CodeGenerator.h"
#include "../Label.h"
#include "../../../intermediate/SymTabEntry.h"
#include "../../../intermediate/symtabimpl/SymTabEntryImpl.h"
#include "../../../intermediate/symtabimpl/Predefined.h"
#include "../../../intermediate/TypeSpec.h"
#include "../../../intermediate/typeimpl/TypeSpecImpl.h"
#include "../../../intermediate/ICodeNode.h"
#include "../../../intermediate/icodeimpl/ICodeNodeImpl.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::intermediate;
using namespace wci::intermediate::symtabimpl;
using namespace wci::intermediate::typeimpl;
using namespace wci::intermediate::icodeimpl;
using namespace wci::backend::compiler;

void CallStandardGenerator::generate(ICodeNode *node)
{
    SymTabEntry *routine_id = node->get_id();
    RoutineCode routine_code = routine_id->get_routine_code();
    ExpressionGenerator expr_generator(this);
    ICodeNode *actual_node = nullptr;

    // Get the actual parameters of the call.
    if (node->get_children()->size() > 0)
    {
        ICodeNode *parms_node = (*node->get_children())[0];
        actual_node = (*parms_node->get_children())[0];
    }

    switch ((RoutineCodeImpl) routine_code)
    {
        case RT_READ:
        case RT_READLN:  generate_read_readln(node, routine_code);
                         break;

        case RT_WRITE:
        case RT_WRITELN: generate_write_writeln(node, routine_code);
                         break;

        case RT_EOF:
        case RT_EOLN:    generate_eof_eoln(node, routine_code);
                         break;

        case RT_ABS:
        case RT_SQR:     generate_abs_sqr(routine_code, actual_node);
                         break;

        case RT_ARCTAN:
        case RT_COS:
        case RT_EXP:
        case RT_LN:
        case RT_SIN:
        case RT_SQRT:    generate_arctan_cos_exp_ln_sin_qqrt(
                                            routine_code, actual_node);
                         break;

        case RT_PRED:
        case RT_SUCC:    generate_pred_succ(routine_code, actual_node);
                         break;

        case RT_CHR:     generate_chr(actual_node); break;
        case RT_ODD:     generate_odd(actual_node); break;
        case RT_ORD:     generate_ord(actual_node); break;

        case RT_ROUND:
        case RT_TRUNC:   generate_round_trunc(routine_code, actual_node);
                         break;

        default:         break;
    }
}

void CallStandardGenerator::generate_read_readln(ICodeNode *call_node,
                                                 RoutineCode routine_code)
{
    ICodeNode *parms_node = call_node->get_children()->size() > 0
                                      ? (*(call_node->get_children()))[0]
                                      : nullptr;
    string *program_name = symtab_stack->get_program_id()->get_name();
    string standard_in_name = *program_name + "/_standardIn";

    if (parms_node != nullptr)
    {
        vector<ICodeNode *> *actuals = parms_node->get_children();

        // Loop to process each actual parameter.
        for (ICodeNode *actual_node : *actuals)
        {
            SymTabEntry *variable_id = actual_node->get_id();
            TypeSpec *actual_type = actual_node->get_typespec();
            TypeSpec *base_type = actual_type->base_type();

            // Generate code to call the appropriate PascalTextIn method.
            emit(GETSTATIC, standard_in_name, "LPascalTextIn;");
            if (base_type == Predefined::integer_type)
            {
                emit(INVOKEVIRTUAL, "PascalTextIn.readInteger()I");
            }
            else if (base_type == Predefined::real_type)
            {
                emit(INVOKEVIRTUAL, "PascalTextIn.readReal()F");
            }
            else if (base_type == Predefined::boolean_type)
            {
                emit(INVOKEVIRTUAL, "PascalTextIn.readBoolean()Z");
            }
            else if (base_type == Predefined::char_type)
            {
                emit(INVOKEVIRTUAL, "PascalTextIn.readChar()C");
            }

            local_stack->increase(1);

            // Store the value that was read into the actual parameter.
            emit_store_variable(variable_id);

            local_stack->decrease(1);
        }
    }

    // READLN: Skip the rest of the input line.
    if (routine_code == (RoutineCode) RT_READLN)
    {
        emit(GETSTATIC, standard_in_name, "LPascalTextIn;");
        emit(INVOKEVIRTUAL, "PascalTextIn.nextLine()V");

        local_stack->use(1);
    }
}

void CallStandardGenerator::generate_eof_eoln(ICodeNode *call_node,
                                              RoutineCode routine_code)
{
    string *program_name = symtab_stack->get_program_id()->get_name();
    string standard_in_name = *program_name + "/_standardIn";

    // Generate code to call the appropriate PascalTextIn method.
    emit(GETSTATIC, standard_in_name, "LPascalTextIn;");
    if (routine_code == (RoutineCode) RT_EOLN)
    {
        emit(INVOKEVIRTUAL, "PascalTextIn.atEoln()Z");
    }
    else {
        emit(INVOKEVIRTUAL, "PascalTextIn.atEof()Z");
    }

    local_stack->increase(1);
}

void CallStandardGenerator::generate_write_writeln(ICodeNode *call_node,
                                                   RoutineCode routine_code)
{
    ICodeNode *parms_node = call_node->get_children()->size() > 0
                              ? (*(call_node->get_children()))[0]
                              : nullptr;
    string buffer = "";
    int expr_count = 0;

    buffer += "\"";

    // There are actual parameters.
    if (parms_node != nullptr)
    {
        vector<ICodeNode *> *actuals = parms_node->get_children();

        // Loop to process each WRITE parameter
        // and build the format string.
        for (ICodeNode *write_parm_node : *actuals)
        {
            vector<ICodeNode *> *children = write_parm_node->get_children();
            ICodeNode *expr_node = (*children)[0];
            ICodeNodeType node_type = expr_node->get_type();

            // Append string constants directly to the format string.
            if (node_type == (ICodeNodeType) NT_STRING_CONSTANT)
            {
                string str = cast((*(expr_node->get_value())), string);
                replace(str, "%", "%%");
                buffer += str;
            }

            // Create and append the appropriate format specification.
            else {
                TypeSpec *typespec = expr_node->get_typespec()->base_type();
                string type_code =
                              typespec->is_pascal_string()         ? "s"
                            : typespec == Predefined::integer_type ? "d"
                            : typespec == Predefined::real_type    ? "f"
                            : typespec == Predefined::boolean_type ? "s"
                            : typespec == Predefined::char_type    ? "c"
                            :                                        "s";

                ++expr_count;  // count the non-constant string parameters
                buffer += "%";

                // Process any field width and precision values.
                Object attribute;
                if (children->size() > 1)
                {
                    Object *value = (*children)[1]->get_value();
                    int width = cast((*value), int);
                    buffer += (width == 0 ? "1" : to_string(width));
                }
                if (children->size() > 2)
                {
                    Object *value = (*children)[2]->get_value();
                    int precision = cast((*value), int);
                    buffer += ".";
                    buffer += (precision == 0 ? "1" : to_string(precision));
                }

                buffer += type_code;
            }
        }

        buffer += (routine_code == (RoutineCode) RT_WRITELN ? "\\n\""
                                                            : "\"");
    }

    emit(GETSTATIC, "java/lang/System/out", "Ljava/io/PrintStream;");
    local_stack->increase(1);

    // WRITELN with no parameters.
    if (parms_node == nullptr)
    {
        emit(INVOKEVIRTUAL, "java/io/PrintStream.println()V");
        local_stack->decrease(1);
    }

    // WRITE or WRITELN with parameters.
    else
    {
        vector<ICodeNode *> *actuals = parms_node->get_children();

        // Load the format string.
        emit(LDC, buffer);
        local_stack->increase(1);

        // Generate code to create the values array for string.format().
        if (expr_count > 0)
        {
            emit_load_constant(expr_count);
            emit(ANEWARRAY, "java/lang/Object");
            local_stack->use(3, 1);

            int index = 0;
            ExpressionGenerator expr_generator =
                                    new ExpressionGenerator(this);

            // Loop to generate code to evaluate each actual parameter.
            for (ICodeNode *write_parm_node : *actuals)
            {
                vector<ICodeNode *> *children = write_parm_node->get_children();
                ICodeNode *expr_node = (*children)[0];
                ICodeNodeType node_type = expr_node->get_type();
                TypeSpec *typespec = expr_node->get_typespec()->base_type();

                // Skip string constants, which were made part of
                // the format string.
                if (node_type != (ICodeNodeType) NT_STRING_CONSTANT)
                {
                    emit(DUP);
                    emit_load_constant(index++);
                    local_stack->increase(2);

                    expr_generator.generate(expr_node);

                    string signature =
                            typespec->get_form() == (TypeForm) TF_SCALAR
                                ? value_of_signature(typespec)
                                : "";

                    // Boolean: Write "true" or "false".
                    if (typespec == Predefined::boolean_type)
                    {
                        Label *true_label = Label::new_label();
                        Label *next_label = Label::new_label();
                        emit(IFNE, true_label);
                        emit(LDC, "\"false\"");
                        emit(GOTO, next_label);
                        emit_label(true_label);
                        emit(LDC, "\"true\"");
                        emit_label(next_label);

                        local_stack->use(1);
                    }

                    // Convert a scalar value to an object.
                    if (signature != "")
                    {
                        emit(INVOKESTATIC, signature);
                    }

                    // Store the value into the values vector.
                    emit(AASTORE);
                    local_stack->decrease(3);
                }
            }

            // Format the string.
            emit(INVOKESTATIC,
                 string("java/lang/String/format(Ljava/lang/String;") +
                 string("[Ljava/lang/Object;)Ljava/lang/String;"));
            local_stack->decrease(2);
        }

        // Print.
        emit(INVOKEVIRTUAL,
             "java/io/PrintStream.print(Ljava/lang/String;)V");
        local_stack->decrease(2);
    }
}

void CallStandardGenerator::generate_abs_sqr(RoutineCode routine_code,
                                             ICodeNode *actual_node)
{
    ExpressionGenerator expr_generator(this);
    expr_generator.generate(actual_node);

    // ABS: Generate code to call the appropriate integer or float
    //      java.lang.Math method.
    // SQR: Multiply the value by itself.
    if (actual_node->get_typespec() == Predefined::integer_type)
    {
        if (routine_code == (RoutineCode) RT_ABS)
        {
            emit(INVOKESTATIC, "java/lang/Math/abs(I)I");
        }
        else
        {
            emit(DUP);
            emit(IMUL);
            local_stack->use(1);
        }
    }
    else
    {
        if (routine_code == (RoutineCode) RT_ABS)
        {
            emit(INVOKESTATIC, "java/lang/Math/abs(F)F");
        }
        else
        {
            emit(DUP);
            emit(FMUL);
            local_stack->use(1);
        }
    }
}

void CallStandardGenerator::generate_arctan_cos_exp_ln_sin_qqrt(
                        RoutineCode routine_code, ICodeNode *actual_node)
{
    string function;
    ExpressionGenerator expr_generator(this);
    expr_generator.generate(actual_node);

    // Convert an integer or real value to double.
    TypeSpec *typespec = actual_node->get_typespec();
    if (typespec == Predefined::integer_type)
    {
        emit(I2D);
    }
    else {
        emit(F2D);
    }

    // Select the appropriate java.lang.Math method.
    switch ((RoutineCodeImpl) routine_code)
    {
        case RT_ARCTAN: function = "atan"; break;
        case RT_COS:    function = "cos";  break;
        case RT_EXP:    function = "exp";  break;
        case RT_SIN:    function = "sin";  break;
        case RT_LN:     function = "log";  break;
        case RT_SQRT:   function = "sqrt"; break;

        default: break;
    }

    // Call the method and convert the result from double to float.
    emit(INVOKESTATIC, "java/lang/Math/" + function + "(D)D");
    emit(D2F);

    local_stack->use(1);
}

void CallStandardGenerator::generate_pred_succ(RoutineCode routine_code,
                                               ICodeNode *actual_node)
{
    ExpressionGenerator expr_generator(this);

    // Generate code to add or subtract 1 from the value.
    expr_generator.generate(actual_node);
    emit(ICONST_1);
    emit(routine_code == (RoutineCode) RT_PRED ? ISUB : IADD);

    local_stack->use(1);
}

void CallStandardGenerator::generate_chr(ICodeNode *actual_node)
{
    ExpressionGenerator expr_generator(this);
    expr_generator.generate(actual_node);
    emit(I2C);
}

void CallStandardGenerator::generate_odd(ICodeNode *actual_node)
{
    ExpressionGenerator expr_generator(this);

    // Generate code to leave the rightmost bit of the value
    // on the operand stack.
    expr_generator.generate(actual_node);
    emit(ICONST_1);
    emit(IAND);

    local_stack->use(1);
}

void CallStandardGenerator::generate_ord(ICodeNode *actual_node)
{
    // A character value is treated as an integer value.
    if (actual_node->get_type() == (ICodeNodeType) NT_STRING_CONSTANT)
    {

        int value = cast((*(actual_node->get_value())), string)[0];
        emit_load_constant(value);
        local_stack->increase(1);
    }
    else
    {
        ExpressionGenerator expr_generator(this);
        expr_generator.generate(actual_node);
    }
}

void CallStandardGenerator::generate_round_trunc(RoutineCode routine_code,
                                                 ICodeNode *actual_node)
{
    ExpressionGenerator expr_generator(this);
    expr_generator.generate(actual_node);

    // ROUND: Generate code to compute floor(value + 0.5).
    if (routine_code == (RoutineCode) RT_ROUND)
    {
        emit_load_constant(0.5f);
        emit(FADD);
        local_stack->use(1);
    }

    // Truncate.
    emit(F2I);
}

}}}}  // namespace wci::backend::compiler::generators
