/**
 * <h1>ExpressionGenerator</h1>
 *
 * <p>Generate code for an expression.</p>
 *
 * <p>Copyright (c) 2017 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#include <string>
#include <vector>
#include <set>
#include "ExpressionGenerator.h"
#include "StatementGenerator.h"
#include "CallGenerator.h"
#include "../CodeGenerator.h"
#include "../../../intermediate/symtabimpl/Predefined.h"
#include "../../../intermediate/ICodeNode.h"
#include "../../../intermediate/icodeimpl/ICodeNodeImpl.h"
#include "../../../intermediate/typeimpl/TypeChecker.h"
#include "../../../intermediate/typeimpl/TypeSpecImpl.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::intermediate;
using namespace wci::intermediate::symtabimpl;
using namespace wci::intermediate::icodeimpl;
using namespace wci::intermediate::typeimpl;
using namespace wci::backend::compiler;

set<ICodeNodeTypeImpl> ExpressionGenerator::ARITH_OPS =
{
    NT_ADD, NT_SUBTRACT, NT_MULTIPLY,
    NT_FLOAT_DIVIDE, NT_INTEGER_DIVIDE, NT_MOD
};

ExpressionGenerator::ExpressionGenerator(CodeGenerator *parent)
    : StatementGenerator(parent)
{
}

void ExpressionGenerator::generate(ICodeNode *node) throw (string)
{
    ICodeNodeTypeImpl node_type = (ICodeNodeTypeImpl) node->get_type();

    switch (node_type)
    {
        case NT_VARIABLE:
        {
            // Generate code to load a variable's value.
            generate_load_value(node);
            break;
        }

        case NT_INTEGER_CONSTANT:
        {
            TypeSpec *typespec = node->get_typespec();
            int value = cast((*(node->get_value())), int);

            // Generate code to load a bool constant
            // 0 (false) or 1 (true).
            if (typespec == Predefined::boolean_type)
            {
                emit_load_constant(value == 1 ? 1 : 0);
            }

            // Generate code to load an integer constant.
            else
            {
                emit_load_constant(value);
            }

            local_stack->increase(1);
            break;
        }

        case NT_REAL_CONSTANT:
        {
            float value = cast((*(node->get_value())), float);

            // Generate code to load a float constant.
            emit_load_constant(value);

            local_stack->increase(1);
            break;
        }

        case NT_STRING_CONSTANT:
        {
            string value = cast((*(node->get_value())), string);

            // Generate code to load a string constant.
            if (node->get_typespec() == Predefined::char_type)
            {
                emit_load_constant(value[0]);
            }
            else
            {
                emit_load_constant(value);
            }

            local_stack->increase(1);
            break;
        }

        case NT_NEGATE:
        {
            // Get the NEGATE node's expression node child.
            vector<ICodeNode *> *children = node->get_children();
            ICodeNode *expr_node = (*children)[0];

            // Generate code to evaluate the expression and
            // negate its value.
            generate(expr_node);
            emit(expr_node->get_typespec() == Predefined::integer_type
                     ? INEG : FNEG);

            break;
        }

        case NT_NOT:
        {
            // Get the NOT node's expression node child.
            vector<ICodeNode *> *children = node->get_children();
            ICodeNode *expr_node = (*children)[0];

            // Generate code to evaluate the expression and NOT its value.
            generate(expr_node);
            emit(ICONST_1);
            emit(IXOR);

            local_stack->use(1);
            break;
        }

        case NT_CALL:
        {
            // Generate code to call a function.
            CallGenerator call_generator(this);
            call_generator.generate(node);

            break;
        }

        // Must be a binary operator.
        default: generate_binary_operator(node, node_type);
    }
}

void ExpressionGenerator::generate_load_value(ICodeNode *variable_node)
{
    vector<ICodeNode *> *variable_children = variable_node->get_children();
    int children_count = variable_children->size();

    // First load the variable's address (structured) or value (scalar).
    TypeSpec *variable_typespec = generate_load_variable(variable_node);

    // Were there any subscript or field modifiers?
    if (children_count > 0)
    {
        ICodeNodeType last_modifier_type =
                        (*variable_children)[children_count-1]->get_type();

        // Array subscript.
        if (last_modifier_type == (ICodeNodeType) NT_SUBSCRIPTS)
        {
            emit_load_array_element(variable_typespec);
            local_stack->decrease(1);
        }

        // Record field.
        else
        {
            emit(INVOKEVIRTUAL,
                 string("java/util/HashMap.get(Ljava/lang/Object;)") +
                 string("Ljava/lang/Object;"));
            emit_check_cast_class(variable_typespec);

            if (!is_structured(variable_typespec))
            {
                emit(INVOKEVIRTUAL, value_signature(variable_typespec));
            }

            local_stack->decrease(1);
        }
    }
}

TypeSpec *ExpressionGenerator::generate_load_variable(
                                                ICodeNode* variable_node)
{
    SymTabEntry *variable_id = variable_node->get_id();
    TypeSpec *variable_typespec = variable_id->get_typespec();
    vector<ICodeNode *> *variable_children = variable_node->get_children();
    int children_count = variable_children->size();

    emit_load_variable(variable_id);
    local_stack->increase(1);

    // Process subscripts and/or fields.
    for (int i = 0; i < children_count; ++i)
    {
        ICodeNode *modifier = (*variable_children)[i];
        ICodeNodeType modifier_type = modifier->get_type();
        bool last = i == children_count-1;

        if (modifier_type == (ICodeNodeType) NT_SUBSCRIPTS)
        {
            variable_typespec =
                generate_array_element(modifier, variable_typespec, last);
        }
        else if (modifier_type == (ICodeNodeType) NT_FIELD)
        {
            variable_typespec = generate_record_field(modifier, last);
        }
    }

    return variable_typespec;
}

TypeSpec *ExpressionGenerator::generate_array_element(
                                            ICodeNode *subscripts_node,
                                            TypeSpec *variable_typespec,
                                            bool last)
{
    vector<ICodeNode *> *subscripts = subscripts_node->get_children();
    ICodeNode *last_subscript = (*subscripts)[subscripts->size()-1];
    TypeSpec *elmt_typespec = variable_typespec;

    for (ICodeNode *subscript : *subscripts)
    {
        generate(subscript);

        TypeSpec *index_typespec = elmt_typespec->get_array_index_type();

        if (index_typespec->get_form() == (TypeForm) TF_SUBRANGE)
        {
            int min = index_typespec->get_subrange_min_value();
            if (min != 0)
            {
                emit_load_constant(min);
                emit(ISUB);
                local_stack->use(1);
            }
        }

        if (!last || (subscript != last_subscript))
        {
            emit(AALOAD);
            local_stack->decrease(1);
        }

        elmt_typespec = elmt_typespec->get_array_elmt_type();
    }

    return elmt_typespec;
}

TypeSpec *ExpressionGenerator::generate_record_field(ICodeNode *field_node,
                                                     bool last)
{
    SymTabEntry *field_id = field_node->get_id();
    string *field_name = field_id->get_name();
    TypeSpec *field_typespec = field_node->get_typespec();

    emit_load_constant(*field_name);
    local_stack->increase(1);

    if (!last)
    {
        emit(INVOKEVIRTUAL,
             string("java/util/HashMap.get(Ljava/lang/Object;)") +
             string("Ljava/lang/Object;"));
        emit_check_cast(field_typespec);
        local_stack->decrease(1);
    }

    return field_typespec;
}

void ExpressionGenerator::generate_binary_operator(
                            ICodeNode *node, ICodeNodeTypeImpl node_type)
{
    // Get the two operand children of the operator node.
    vector<ICodeNode *> *children = node->get_children();
    ICodeNode *operand_node1 = (*children)[0];
    ICodeNode *operand_node2 = (*children)[1];
    TypeSpec *typespec1 = operand_node1->get_typespec();
    TypeSpec *typespec2 = operand_node2->get_typespec();

    bool integer_mode =
                   TypeChecker::are_both_integer(typespec1, typespec2)
                || (typespec1->get_form() == (TypeForm) TF_ENUMERATION)
                || (typespec2->get_form() == (TypeForm) TF_ENUMERATION);
    bool real_mode =
                   TypeChecker::is_at_least_one_real(typespec1, typespec2)
                || (node_type == NT_FLOAT_DIVIDE);
    bool character_mode =    TypeChecker::is_char(typespec1)
                          && TypeChecker::is_char(typespec2);
    bool string_mode =    typespec1->is_pascal_string()
                       && typespec2->is_pascal_string();

    if (!string_mode)
    {
        // Emit code to evaluate the first operand.
        generate(operand_node1);
        if (real_mode && TypeChecker::is_integer(typespec1)) emit(I2F);

        // Emit code to evaluate the second operand.
        generate(operand_node2);
        if (real_mode && TypeChecker::is_integer(typespec2)) emit(I2F);
    }

    // ====================
    // Arithmetic operators
    // ====================

    if (ARITH_OPS.find(node_type) != ARITH_OPS.end())
    {
        if (integer_mode)
        {
            // Integer operations.
            switch (node_type)
            {
                case NT_ADD:            emit(IADD); break;
                case NT_SUBTRACT:       emit(ISUB); break;
                case NT_MULTIPLY:       emit(IMUL); break;
                case NT_FLOAT_DIVIDE:   emit(FDIV); break;
                case NT_INTEGER_DIVIDE: emit(IDIV); break;
                case NT_MOD:            emit(IREM); break;

                default: break;  // should never get here
            }
        }
        else
        {
            // Float operations.
            switch (node_type)
            {
                case NT_ADD:          emit(FADD); break;
                case NT_SUBTRACT:     emit(FSUB); break;
                case NT_MULTIPLY:     emit(FMUL); break;
                case NT_FLOAT_DIVIDE: emit(FDIV); break;

                default: break;  // should never get here
            }
        }

        local_stack->decrease(1);
    }

    // ==========
    // AND and OR
    // ==========

    else if (node_type == NT_AND)
    {
        emit(IAND);
        local_stack->decrease(1);
    }
    else if (node_type == NT_OR)
    {
        emit(IOR);
        local_stack->decrease(1);
    }

    // ====================
    // Relational operators
    // ====================

    else
    {
        Label *true_label = Label::new_label();
        Label *next_label = Label::new_label();

        if (integer_mode || character_mode)
        {
            switch (node_type)
            {
                case NT_EQ: emit(IF_ICMPEQ, true_label); break;
                case NT_NE: emit(IF_ICMPNE, true_label); break;
                case NT_LT: emit(IF_ICMPLT, true_label); break;
                case NT_LE: emit(IF_ICMPLE, true_label); break;
                case NT_GT: emit(IF_ICMPGT, true_label); break;
                case NT_GE: emit(IF_ICMPGE, true_label); break;

                default: break;  // should never get here
            }

            local_stack->decrease(2);
        }

        else if (real_mode)
        {
            emit(FCMPG);

            switch (node_type)
            {
                case NT_EQ: emit(IFEQ, true_label); break;
                case NT_NE: emit(IFNE, true_label); break;
                case NT_LT: emit(IFLT, true_label); break;
                case NT_LE: emit(IFLE, true_label); break;
                case NT_GT: emit(IFGT, true_label); break;
                case NT_GE: emit(IFGE, true_label); break;

                default: break;  // should never get here
            }

            local_stack->decrease(2);
        }

        else if (string_mode)
        {
            // Load the value of the first string operand.
            generate(operand_node1);
            if (operand_node1->get_type() !=
                                    (ICodeNodeType) NT_STRING_CONSTANT)
            {
                emit(INVOKEVIRTUAL,
                     string("java/lang/StringBuilder.toString()") +
                     string("Ljava/lang/String;"));
            }

            // Load the value of the second string operand.
            generate(operand_node2);
            if (operand_node2->get_type() !=
                                    (ICodeNodeType) NT_STRING_CONSTANT)
            {
                emit(INVOKEVIRTUAL,
                     string("java/lang/StringBuilder.toString()") +
                     string("Ljava/lang/String;"));
            }

            emit(INVOKEVIRTUAL,
                 "java/lang/String.compareTo(Ljava/lang/String;)I");

            switch (node_type)
            {
                case NT_EQ: emit(IFEQ, true_label); break;
                case NT_NE: emit(IFNE, true_label); break;
                case NT_LT: emit(IFLT, true_label); break;
                case NT_LE: emit(IFLE, true_label); break;
                case NT_GT: emit(IFGT, true_label); break;
                case NT_GE: emit(IFGE, true_label); break;

                default: break;  // should never get here
            }

            local_stack->decrease(2);
        }

        emit(ICONST_0); // false
        emit(GOTO, next_label);
        emit_label(true_label);
        emit(ICONST_1); // true
        emit_label(next_label);

        local_stack->increase(1);
    }
}

}}}}  // namespace wci::backend::compiler::generators
