/**
 * <h1>SelectGenerator</h1>
 *
 * <p>Generate code for a select statement.</p>
 *
 * <p>Copyright (c) 2018 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#include <string>
#include <vector>
#include "SelectGenerator.h"
#include "StatementGenerator.h"
#include "ExpressionGenerator.h"
#include "../CodeGenerator.h"
#include "../Label.h"
#include "../../../intermediate/ICodeNode.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::intermediate;
using namespace wci::backend::compiler;

SelectGenerator::SelectGenerator(CodeGenerator *parent)
    : StatementGenerator(parent)
{
}

void SelectGenerator::generate(ICodeNode *node) throw (string)
{
    vector<ICodeNode *> *select_children = node->get_children();
    ICodeNode *expr_node = (*select_children)[0];
    vector<Label *> branch_labels;
    vector<ValueLabelPair> pairs;

    branch_labels.push_back(Label::new_label());  // "next" label

    // Generate code to evaluate the SELECT expression.
    ExpressionGenerator expr_generator(this);
    expr_generator.generate(expr_node);

    // Process the select branches.
    process_select_branches(*select_children, branch_labels, pairs);

    // Generate code for the LOOKUPSWITCH and the branch statements.
    generate_lookup_switch(pairs, branch_labels);
    generate_branch_statements(*select_children, branch_labels);
}

void SelectGenerator::process_select_branches(
                                    vector<ICodeNode *>& select_children,
                                    vector<Label *>& branch_labels,
                                    vector<ValueLabelPair>& pairs)
{
    // Loop over the SELECT branches.
    for (int i = 1; i < select_children.size(); i++)
    {
        ICodeNode *branch_node = select_children[i];
        ICodeNode *constants_node = (*(branch_node->get_children()))[0];
        Label *branch_label = Label::new_label();

        branch_labels.push_back(branch_label);

        // Loop over the constants children of the branch's
        // SELECT_CONSTANTS node and create the label-value pairs.
        vector<ICodeNode *> *constants_list = constants_node->get_children();
        for (ICodeNode *constant_node : *constants_list)
        {
            Object value = *(constant_node->get_value());
            int i = constant_node->get_type()
                                    == (ICodeNodeType) NT_STRING_CONSTANT
                                               ? cast(value, string)[0]
                                               : cast(value, int);

            pairs.push_back(ValueLabelPair(i, branch_label));
        }
    }

    // Sort and return the list of value-label pairs.
    sort_pairs(pairs);
}

void SelectGenerator::sort_pairs(vector<ValueLabelPair>& pairs)
{
    // Simple exchange sort on the SELECT branch values.
    for (int i = 0; i < pairs.size()-1; i++)
    {
        for (int j = i+1; j < pairs.size(); j++)
        {
            if (pairs[i].value > pairs[j].value)
            {
                ValueLabelPair temp = pairs[i];
                pairs[i] = pairs[j];
                pairs[j] = temp;
            }
        }
    }
}

void SelectGenerator::generate_lookup_switch(vector<ValueLabelPair>& entries,
                                             vector<Label *>& branch_labels)
{
    emit_blank_line();
    emit(LOOKUPSWITCH);

    // For each entry, emit the value and its label.
    for (ValueLabelPair entry : entries)
    {
        emit_label(entry.value, entry.label);
    }

    // Emit the default label;
    emit_label("default", branch_labels[0]);

    local_stack->decrease(1);
}

void SelectGenerator::generate_branch_statements(
                                    vector<ICodeNode *>& select_children,
                                    vector<Label *>& branch_labels)
    throw (string)
{
    StatementGenerator stmt_generator(this);
    emit_blank_line();

    // Loop to emit each branch label and then generate code for
    // the corresponding branch statement.
    for (int i = 1; i < select_children.size(); i++)
    {
        ICodeNode *branch_node = select_children[i];
        ICodeNode *statement_node = (*(branch_node->get_children()))[1];

        emit_label(branch_labels[i]);
        stmt_generator.generate(statement_node);

        // Branch to the "next" label.
        emit(GOTO, branch_labels[0]);
    }

    // Emit the "next" label.
    emit_label(branch_labels[0]);
    emit_blank_line();
}

}}}}  // namespace wci::backend::compiler::generators
