/**
 * <h1>SelectGenerator</h1>
 *
 * <p>Generate code for a select statement.</p>
 *
 * <p>Copyright (c) 2018 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#ifndef WCI_BACKEND_COMPILER_GENERATORS_SELECTGENERATOR_H_
#define WCI_BACKEND_COMPILER_GENERATORS_SELECTGENERATOR_H_

#include <string>
#include <vector>
#include "StatementGenerator.h"
#include "../CodeGenerator.h"
#include "../Label.h"
#include "../../../intermediate/ICodeNode.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::intermediate;
using namespace wci::backend::compiler;

/**
 * A value-label pair for the LOOKUPSWITCH: A SELECT branch value
 *                                          and its statement label.
 */
class ValueLabelPair
{
public:
    /**
     * Constructor.
     * @param value the SELECT branch value.
     * @param label the corresponding statement label.
     */
    ValueLabelPair(int value, Label *label)
        : value(value), label(label)
    {
    }

    int value;
    Label *label;
};

class SelectGenerator : public StatementGenerator
{
public:
    /**
     * Constructor.
     * @param the parent executor.
     */
    SelectGenerator(CodeGenerator *parent);

    /**
     * Generate code for a SELECT statement.
     * @param node the root node of the statement.
     * @throw a string message if an error occurred.
     */
    void generate(ICodeNode *node) throw (string);

private:
    /**
     * Process the SELECT branches.
     * @param select_children the child nodes of the SELECT node.
     * @param branch_labels the branch labels.
     * @param pairs the table entries.
     */
    void process_select_branches(vector<ICodeNode *>& select_children,
                                 vector<Label *>& branch_labels,
                                 vector<ValueLabelPair>& pairs);

    /**
     * Sort the label-value pairs into ascending order by value.
     * @param pairs the list of pairs.
     */
    void sort_pairs(vector<ValueLabelPair>& pairs);

    /**
     * Generate code for the LOOKUPSWITCH instruction.
     * @param entries the table entries.
     * @param branch_labels the branch labels.
     */
    void generate_lookup_switch(vector<ValueLabelPair>& entries,
                                vector<Label *>& branch_labels);

    /**
     * Generate code for the branch statements.
     * @param select_children the child nodes of the SELECT node.
     * @param branch_labels the branch labels
     * @throw a string message if an error occurred.
     */
    void generate_branch_statements(vector<ICodeNode *>& select_children,
                                    vector<Label *>& branch_labels)
        throw (string);
};

}}}}  // namespace wci::backend::compiler::generators

#endif /* WCI_BACKEND_COMPILER_GENERATORS_SELECTGENERATOR_H_ */
