/**
 * <h1>StructuredDataGenerator</h1>
 *
 * <p>Generate code to allocate arrays, records, and strings.</p>
 *
 * <p>Copyright (c) 2017 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#include <string>
#include <vector>
#include "StructuredDataGenerator.h"
#include "../CodeGenerator.h"
#include "../../BackendFactory.h"
#include "../../../intermediate/SymTabEntry.h"
#include "../../../intermediate/symtabimpl/SymTabEntryImpl.h"
#include "../../../intermediate/symtabimpl/Predefined.h"
#include "../../../intermediate/TypeSpec.h"
#include "../../../intermediate/typeimpl/TypeSpecImpl.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

using namespace std;
using namespace wci::backend;
using namespace wci::intermediate::symtabimpl;
using namespace wci::intermediate::typeimpl;

void StructuredDataGenerator::generate(SymTabEntry *routine_id)
    throw (string)
{
    SymTab *symtab = routine_id->get_routine_symtab();
    vector<SymTabEntry *> ids = symtab->sorted_entries();

    // Loop over all the symbol table's identifiers to generate
    // data allocation code for string variables.
    emit_blank_line();
    for (SymTabEntry *id : ids)
    {
        if (id->get_definition() == (Definition) DF_VARIABLE)
        {
            TypeSpec *id_typespec = id->get_typespec();

            if (is_structured(id_typespec))
            {
                TypeForm id_form = id_typespec->get_form();

                if (   id_typespec->is_pascal_string()
                    || (id_form == (TypeForm) TF_RECORD))
                {
                    generate_allocate_data(id, id_typespec);
                }
                else
                {
                    generate_allocate_array(id, id_typespec);
                }
            }
        }
    }
}

void StructuredDataGenerator::generate_allocate_data(
                                                SymTabEntry *variable_id,
                                                TypeSpec *data_typespec)
{
    TypeForm form = data_typespec->get_form();

    if (data_typespec->is_pascal_string())
    {
        generate_allocate_string(variable_id, data_typespec);
    }
    else if (form == (TypeForm) TF_ARRAY)
    {
        generate_allocate_elements(variable_id, data_typespec, 1);
    }
    else if (form == (TypeForm) TF_RECORD)
    {
        generate_allocate_record(variable_id, data_typespec);
    }
}

void StructuredDataGenerator::generate_allocate_array(
                                            SymTabEntry *variable_id,
                                            TypeSpec *array_typespec)
{
    TypeSpec *elmt_typespec = array_typespec;
    int dimension_count = 0;

    // Count the dimensions and emit a load constant of each size.
    emit_blank_line();
    do {
        int size = elmt_typespec->get_array_elmt_count();
        ++dimension_count;
        emit_load_constant(size);
        elmt_typespec = elmt_typespec->get_array_elmt_type();
    } while (   (elmt_typespec->get_form() == (TypeForm) TF_ARRAY)
             && (!elmt_typespec->is_pascal_string()));

    // The array element type.
    elmt_typespec = elmt_typespec->base_type();
    TypeForm elmt_form = elmt_typespec->get_form();
    string type_name =
          elmt_typespec == Predefined::integer_type ? "int"
        : elmt_typespec == Predefined::real_type    ? "float"
        : elmt_typespec == Predefined::boolean_type ? "boolean"
        : elmt_typespec == Predefined::char_type    ? "char"
        : elmt_form == (TypeForm) TF_ENUMERATION    ? "int"
        : elmt_form == (TypeForm) TF_RECORD         ? "java/util/HashMap"
        : elmt_typespec->is_pascal_string()         ? "java/lang/StringBuilder"
        :                                             "";

    // One-dimensional array.
    if (dimension_count == 1)
    {
        OpCode new_array = is_structured(elmt_typespec) ? ANEWARRAY
                                                        : NEWARRAY;
        emit(new_array, type_name);
    }

    // Multidimensional array.
    else
    {
        emit(MULTIANEWARRAY, type_descriptor(variable_id->get_typespec()),
             to_string(dimension_count));
    }

    local_stack->use(dimension_count+1, dimension_count);

    // Allocate data for structured elements.
    if (is_structured(elmt_typespec))
    {
        generate_allocate_data(variable_id, variable_id->get_typespec());
    }

    // Store the allocation into the array variable.
    generate_store_data(variable_id);
}

void StructuredDataGenerator::generate_allocate_elements(
                                            SymTabEntry *variable_id,
                                            TypeSpec *elmt_typespec,
                                            int dimension_index)
{
    int count = elmt_typespec->get_array_elmt_count();
    int temp_index = local_variables->reserve();  // temporary loop variable
    Label *loop_start_label = Label::new_label();
    Label *loop_exit_label  = Label::new_label();

    // Initialize temporary variable to 0.
    emit_blank_line();
    emit_load_constant(0);
    emit_store_local(Predefined::integer_type, temp_index);

    // Top of loop: Compare the temporary variable to the element count.
    emit_label(loop_start_label);
    emit_load_local(Predefined::integer_type, temp_index);
    emit_load_constant(count);
    emit(IF_ICMPGE, loop_exit_label);

    emit_blank_line();
    emit(DUP);  // duplicate dimension address for use by each element

    local_stack->use(4, 3);
    elmt_typespec = elmt_typespec->get_array_elmt_type();

    // Allocate data for a string or record element.
    if (   elmt_typespec->is_pascal_string()
        || (elmt_typespec->get_form() == (TypeForm) TF_RECORD))
    {
        emit_load_local(Predefined::integer_type, temp_index);
        local_stack->increase(1);

        generate_allocate_data(nullptr, elmt_typespec);
    }

    // Allocate data for the next array dimension.
    else
    {
        emit_load_local(Predefined::integer_type, temp_index);
        emit(AALOAD);
        local_stack->use(2);

        generate_allocate_elements(variable_id, elmt_typespec,
                                   dimension_index+1);
    }

    // Bottom of loop: Increment the temporary variable and
    //                 branch back to the top of the loop.
    emit(IINC, temp_index, 1);
    emit(GOTO, loop_start_label);
    emit_label(loop_exit_label);

    // Pop off the extraneous duplicated dimension address.
    if (dimension_index > 1)
    {
        emit(POP);
        local_stack->decrease(1);
    }

    local_variables->release(temp_index);
}

const string StructuredDataGenerator::PADDED_STRING_CREATE =
    "PaddedString.create(I)Ljava/lang/StringBuilder;";

void StructuredDataGenerator::generate_allocate_string(
                                            SymTabEntry *variable_id,
                                            TypeSpec *string_typespec)
{
    int length = string_typespec->get_array_elmt_count();

    // Allocate a blank-filled string of the correct length.
    emit_load_constant(length);
    emit(INVOKESTATIC, PADDED_STRING_CREATE);
    local_stack->increase(1);

    // Store the allocation into the string variable.
    generate_store_data(variable_id);
}

void StructuredDataGenerator::generate_allocate_record(
                                            SymTabEntry *variable_id,
                                            TypeSpec *record_typespec)
{
    SymTab *record_symtab = record_typespec->get_record_symtab();

    // Allocate a hash map.
    emit_blank_line();
    emit(NEW, "java/util/HashMap");
    emit(DUP);
    emit(INVOKENONVIRTUAL, "java/util/HashMap/<init>()V");
    local_stack->use(2, 1);

    // Allocate the record fields.
    generate_allocate_fields(record_symtab);

    // Store the allocation into the record variable.
    generate_store_data(variable_id);
}

void StructuredDataGenerator::generate_allocate_fields(SymTab *symtab)
{
    vector<SymTabEntry *> fields = symtab->sorted_entries();

    // Loop over all the symbol table's field identifiers
    // to generate data allocation code.
    for (SymTabEntry *field_id : fields)
    {
        TypeSpec *id_typespec = field_id->get_typespec()->base_type();

        // Load the field name.
        emit(DUP);
        emit_load_constant(*(field_id->get_name()));
        local_stack->increase(2);

        // Allocate data for a string, array, or record field.
        if (is_structured(id_typespec)) {
            TypeForm id_form = id_typespec->get_form();

            if (   id_typespec->is_pascal_string()
                || (id_form == (TypeForm) TF_RECORD))
            {
                generate_allocate_data(field_id, id_typespec);
            }
            else {
                generate_allocate_array(field_id, id_typespec);
            }
        }

        // Initialize a scalar field.
        else
        {
            Object value = BackendFactory::default_value(id_typespec);

            if (id_typespec == Predefined::integer_type)
            {
                emit_load_constant(cast(value, int));
            }
            else if (id_typespec == Predefined::real_type)
            {
                emit_load_constant(cast(value, float));
            }
            else if (id_typespec == Predefined::boolean_type)
            {
                emit_load_constant(cast(value, bool) ? 1 : 0);
            }
            else /* id_typespec == Predefined.charType */
            {
                emit_load_constant(cast(value, char));
            }

            // Promote a scalar value to an object.
            emit(INVOKESTATIC, value_of_signature(id_typespec));
            local_stack->increase(1);

            generate_store_data(field_id);
        }
    }
}

void StructuredDataGenerator::generate_store_data(SymTabEntry *variable_id)
{
    // Store into an array element.
    if (variable_id == nullptr)
    {
        emit(AASTORE);
        local_stack->decrease(3);
    }

    // Store into a record field.
    else if (variable_id->get_definition() == (Definition) DF_FIELD)
    {
        emit(INVOKEVIRTUAL,
             string("java/util/HashMap.put(Ljava/lang/Object;Ljava/lang/Object;)") +
             string("Ljava/lang/Object;"));
        emit(POP);

        local_stack->decrease(3);
    }

    // Store into a variable.
    else
    {
        emit_store_variable(variable_id);
        local_stack->decrease(1);
    }
}

}}}}  // namespace wci::backend::compiler::generators
