/**
 * <h1>StructuredDataGenerator</h1>
 *
 * <p>Generate code to allocate arrays, records, and strings.</p>
 *
 * <p>Copyright (c) 2017 by Ronald Mak</p>
 * <p>For instructional purposes only.  No warranties.</p>
 */
#ifndef STRUCTUREDDATAGENERATOR_H_
#define STRUCTUREDDATAGENERATOR_H_

#include "../CodeGenerator.h"
#include "../../../intermediate/SymTabEntry.h"

namespace wci { namespace backend { namespace compiler { namespace generators {

class StructuredDataGenerator : public CodeGenerator
{
public:
    /**
     * Constructor.
     * @param the parent generator.
     */
    StructuredDataGenerator(CodeGenerator *parent) : CodeGenerator(parent) {}

    /**
     * Generate code to allocate the structured data of a program,
     * procedure, or function.
     * @param routine_id the symbol table entry of the routine's name.
     * @throw a string message if an error occurred.
     */
    void generate(SymTabEntry *routine_id) throw (string);

    /**
     * Generate code to allocate array and record variables and fields.
     * @param symtab the record's symbol table containing fields.
     */
    void generate_allocate_fields(SymTab *symtab);

private:
    /**
     * Generate data to allocate a string, array, or record variable.
     * @param variable_id the variable's symbol table entry.
     * @param typespec the variable's data type.
     */
    void generate_allocate_data(SymTabEntry *variable_id,
                                TypeSpec *typespec);

    /**
     * Generate code to allocate data for an array variable.
     * @param variable_id the symbol table entry of the variable.
     * @param array_typespec the array type.
     */
    void generate_allocate_array(SymTabEntry *variable_id,
                                 TypeSpec *array_typespec);

    /**
     * Generate code that loops over the array dimensions to allocate
     * data for each element.
     * @param variable_id the symbol table entry of the variable.
     * @param elmt_typespec the data type of the array element.
     * @param dimension_index the first dimension is 1, second is 2, etc.
     */
    void generate_allocate_elements(SymTabEntry *variable_id,
                                    TypeSpec *elmt_typespec,
                                    int dimension_index);

    static const string PADDED_STRING_CREATE;

    /**
     * Generate code to allocate a string variable as a StringBuilder.
     * @param variable_id the symbol table entry of the variable.
     * @param string_typespec the string data type.
     */
    void generate_allocate_string(SymTabEntry *variable_id,
                                  TypeSpec *string_typespec);

    /**
     * Generate code to allocate a record variable as a HashMap.
     * @param variable_id the symbol table entry of the variable.
     * @param record_typespec the record data type.
     */
    void generate_allocate_record(SymTabEntry *variable_id,
                                  TypeSpec *record_typespec);

    /**
     * Generate code to store allocated data into a variable,
     * record, field, or array element.
     * @param variable_id the variable's symbol table entry,
     *                    or nullptr to store into an array element.
     */
    void generate_store_data(SymTabEntry *variable_id);
};

}}}}  // namespace wci::backend::compiler::generators

#endif /* STRUCTUREDDATAGENERATOR_H_ */
